const wchar_t *about_text = L"\
PDF Compatibility viewer X-Tension\
\n\n\
This viewer X-Tension provides a workaround for the 'disappearing fonts during printing' \
bug in the original Oracle Outside In Viewer component. The bug was introduced in the \
viewer version 8.52 and as of now is still persistent in the latest version (8.53). \
\n\n\
The X-Tension utilizes the GhostScript API through gsdll64.dll and passes the result of \
the following conversion command back to X-Ways:\
\n\n\
gs -sDEVICE=pdfwrite -dNoOutputFonts -dCompatibilityLevel=1.5 -r150 -dQUIET -dSAFER -o <output file> <input file>";

#include <windows.h>
#include "main.h"

EXPORT LONG XT_Init( DWORD nVersion, DWORD nFlags, HANDLE hMainWnd, struct LicenseInfo* LicInfo )
{
	if ( !InitXWaysApi() )
	{
		return -1;
	}

#ifdef DEBUG
	XWF_OutputMessage( L"XT_Init() called", 2 );
#endif

	if ( !InitGhostScriptApi() )
	{
		XWF_OutputMessage( L"[ERROR] Could not load the GhostScript API (gsdll64.dll).", 0 );
		return -1;
	}

	XWF_OutputMessage( L"PDF Compatibility viewer X-Tension loaded.", 2 );

	return 2; //thread-safe X-Tension
}

EXPORT PVOID XT_View( HANDLE hItem, LONG nItemID, HANDLE hVolume, HANDLE hEvidence, PVOID lpReserved, PINT64 nResSize )
{
#ifdef DEBUG
	XWF_OutputMessage( L"------------------------------------------------------", 2 );
	XWF_OutputMessage( L"XT_View() called", 2 );
#endif

	*nResSize = -1; //'not responsible for this type of file'

	//Check filesize
	const INT64 file_size = XWF_GetSize( hItem, (LPVOID) 1 );
	if ( 150 > file_size )
	{
		//too small for a valid pdf file
		return NULL;
	}

	//Check general file signature
	char sig_buf[5] = {};
	const DWORD sig_br = XWF_Read( hItem, 0, sig_buf, 4 );
	if ( 4 != sig_br || 0 != strcmp( "%PDF", sig_buf ) )
	{
		//not enough bytes read or not a pdf signature
		return NULL;
	}

	*nResSize = -2; //'an error occured'

	//Prepare two temporary files for converting
	wchar_t tmp_path[MAX_PATH];
	DWORD tmp_path_res = GetTempPath( MAX_PATH, tmp_path );
	if ( 0 == tmp_path_res || MAX_PATH < tmp_path_res )
	{
		return NULL;
	}
	wchar_t tmp_src_filename[MAX_PATH];
	wchar_t tmp_dst_filename[MAX_PATH];
	if ( 0 == GetTempFileName( tmp_path, L"KGB", 0, tmp_src_filename )
		|| 0 == GetTempFileName( tmp_path, L"KGB", 0, tmp_dst_filename ) )
	{
		return NULL;
	}
	HANDLE tmp_src_file = CreateFile( tmp_src_filename, GENERIC_WRITE, 0,
		NULL, OPEN_EXISTING, FILE_ATTRIBUTE_TEMPORARY, NULL );
	if ( INVALID_HANDLE_VALUE == tmp_src_file )
	{
		DeleteFile( tmp_src_filename );
		DeleteFile( tmp_dst_filename );
		return NULL;
	}
#ifdef DEBUG
	XWF_OutputMessage( L"\tTemporary file: ", 2 );
	XWF_OutputMessage( tmp_src_filename, 3 );
	XWF_OutputMessage( L"\tTemporary file: ", 2 );
	XWF_OutputMessage( tmp_dst_filename, 3 );
#endif

	HANDLE heap = GetProcessHeap();

	//Read file contents from X-Ways
	char *file_buf = (char*) HeapAlloc( heap, 0, file_size );
	if ( NULL == file_buf )
	{
		DeleteFile( tmp_src_filename );
		DeleteFile( tmp_dst_filename );
		return NULL;
	}
	const DWORD file_br = XWF_Read( hItem, 0, file_buf, (DWORD) file_size );
	if ( file_size != file_br )
	{
		//not enough bytes read
		HeapFree( heap, 0, file_buf );
		DeleteFile( tmp_src_filename );
		DeleteFile( tmp_dst_filename );
		return NULL;
	}
#ifdef DEBUG
	else XWF_OutputMessage( L"\tXWF_Read() successful", 2 );
#endif

	//Write file contents into the 'source' temp file
	DWORD bytes_written = 0;
	const BOOL write_res = WriteFile( tmp_src_file, file_buf, (DWORD) file_size, &bytes_written, NULL );
	CloseHandle( tmp_src_file );
	if ( FALSE == write_res || file_size != bytes_written )
	{
		HeapFree( heap, 0, file_buf );
		DeleteFile( tmp_src_filename );
		DeleteFile( tmp_dst_filename );
		return NULL;
	}
#ifdef DEBUG
	else XWF_OutputMessage( L"\tWriteFile() successful", 2 );
#endif

	//Original file buffer isn't needed anymore
	HeapFree( heap, 0, file_buf );

	//Initialize and run the GhostScript API on the temp files
	bool gs_was_successful = false;

	void *pinstance = NULL;
	void *caller_handle = NULL;
	if ( 0 == gsapi_new_instance( &pinstance, caller_handle ) )
	{
		/*
		It should be possible to pass wchar_t *argv to gsapi_init_with_argsW(), eventually
		using gsapi_set_arg_encoding(instance, 1). For some reason all I got in return from
		gsapi_init_with_args() was either a -100 (gs_error_Fatal) or a segfault :(
		Let's use WideCharToMultiByte() instead.
		*/
		//Convert paths of temporary files to char strings
		const int src_mb_size = WideCharToMultiByte( CP_UTF8, 0, tmp_src_filename, -1, NULL, 0, NULL, NULL );
		char *tmp_src_filename_mb = (char*) HeapAlloc( heap, 0, src_mb_size );
		WideCharToMultiByte( CP_UTF8, 0, tmp_src_filename, -1, tmp_src_filename_mb, src_mb_size, NULL, NULL );
		const int dst_mb_size = WideCharToMultiByte( CP_UTF8, 0, tmp_dst_filename, -1, NULL, 0, NULL, NULL );
		char *tmp_dst_filename_mb = (char*) HeapAlloc( heap, 0, dst_mb_size );
		WideCharToMultiByte( CP_UTF8, 0, tmp_dst_filename, -1, tmp_dst_filename_mb, dst_mb_size, NULL, NULL );

		//Set arguments for the GhostScript call
		int argc = 10;
		char *argv[10];
		argv[0] = "1337";                     //first element is ignored
		argv[1] = "-sDEVICE=pdfwrite";        //creates a PDF file
		argv[2] = "-dNoOutputFonts";          //IMPORTANT: fixes font bug
		argv[3] = "-dCompatibilityLevel=1.5"; //PDF version of the output file
		argv[4] = "-r150";                    //DPI resolution
		argv[5] = "-dQUIET";                  //disable comments on stdout
		argv[6] = "-dSAFER";                  //restrict file operations
		argv[7] = "-o";                       //includes -dBATCH and -dNOPAUSE
		argv[8] = tmp_dst_filename_mb;
		argv[9] = tmp_src_filename_mb;

		//A successful GhostScript call returns 0
		gs_was_successful = ( 0 == gsapi_init_with_args( pinstance, argc, argv ) );

		HeapFree( heap, 0, tmp_src_filename_mb );
		HeapFree( heap, 0, tmp_dst_filename_mb );
	}

	//The first temporary file can go now
	DeleteFile( tmp_src_filename );

	//Close the GhostScript instance
	if ( 0 != gsapi_exit( pinstance ) )
	{
		gs_was_successful = false;
	}
	gsapi_delete_instance( pinstance );

	//Abort if anything went wrong with the GhostScript API
	if ( !gs_was_successful )
	{
		XWF_OutputMessage( L"[ERROR] GhostScript conversion failed.", 2 );
		DeleteFile( tmp_dst_filename );
		return NULL;
	}
#ifdef DEBUG
	else XWF_OutputMessage( L"\tGhostScript conversion successful", 2 );
#endif

	//Read the 'destination' temporary file into the buffer
	//Open the file
	HANDLE tmp_dst_file = CreateFile( tmp_dst_filename, GENERIC_READ, 0,
		NULL, OPEN_EXISTING, FILE_ATTRIBUTE_TEMPORARY, NULL );
	if ( INVALID_HANDLE_VALUE == tmp_dst_file )
	{
		DeleteFile( tmp_dst_filename );
		return NULL;
	}
	//Prepare a new buffer with matching size
	const DWORD tmp_dst_filesize = GetFileSize( tmp_dst_file, NULL );
	file_buf = (char*) HeapAlloc( heap, 0, tmp_dst_filesize );
	if ( NULL == file_buf )
	{
		DeleteFile( tmp_dst_filename );
		return NULL;
	}
	//Read file into buffer
	DWORD bytes_read = 0;
	const BOOL read_res = ReadFile( tmp_dst_file, file_buf, tmp_dst_filesize, &bytes_read, NULL );
	//Either way, get rid of the other temporary file, too
	CloseHandle( tmp_dst_file );
	DeleteFile( tmp_dst_filename );
	if ( FALSE == read_res || tmp_dst_filesize != bytes_read )
	{
		HeapFree( heap, 0, file_buf );
		return NULL;
	}
#ifdef DEBUG
	else XWF_OutputMessage( L"\tReadFile() successful", 2 );
	XWF_OutputMessage( L"XT_View() completed successfully, buffer size is", 2 );
	wchar_t buf[32];
	wsprintf( buf, L"%d", bytes_read );
	XWF_OutputMessage( buf, 3 );
#endif

	*nResSize = bytes_read;

	return file_buf;
}

EXPORT BOOL XT_ReleaseMem( PVOID lpBuffer )
{
#ifdef DEBUG
	XWF_OutputMessage( L"XT_ReleaseMem() called", 2 );
#endif

	if ( NULL == lpBuffer )
	{
		return FALSE;
	}

	HeapFree( GetProcessHeap(), 0, lpBuffer );

	return TRUE;
}

EXPORT LONG XT_About( HANDLE hParentWnd, PVOID lpReserved )
{
	MessageBox( NULL, about_text, L"About", MB_ICONINFORMATION );
	return 0;
}
